{
 "cells": [
  {
   "cell_type": "markdown",
   "metadata": {
    "collapsed": false
   },
   "source": [
    "# Iterated Prisoner's Dilemma"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {
    "collapsed": false
   },
   "source": [
    "Game class for 2-player Iterated Prisoner's Dilemma. One can modify the number of rounds and the miscommunication chance."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 1,
   "metadata": {
    "collapsed": false
   },
   "outputs": [
   ],
   "source": [
    "import numpy as np\n",
    "\n",
    "class game:\n",
    "    def __init__(self,player0,player1):\n",
    "        self.player0 = player0\n",
    "        self.player1 = player1\n",
    "        self.decisions = []\n",
    "        self.scores = []\n",
    "\n",
    "    def play(self,n=10,m=0,p=1):\n",
    "        misc = lambda x: int(not x) if np.random.random() <= m/100. else x\n",
    "        for i in range(n):\n",
    "            newDecisions = [misc(self.player0.decision(0,self.decisions)),misc(self.player1.decision(1,self.decisions))]\n",
    "            self.decisions.append(newDecisions)\n",
    "            if newDecisions == [0,0]:\n",
    "                newScores = [0,0]\n",
    "            elif newDecisions == [1,0]:\n",
    "                newScores = [-1,3]\n",
    "            elif newDecisions == [0,1]:\n",
    "                newScores = [3,-1]\n",
    "            else:\n",
    "                newScores = [2,2]\n",
    "            self.scores.append(newScores)\n",
    "        player0Score = sum([a for a,b in self.scores])\n",
    "        player1Score = sum([b for a,b in self.scores])\n",
    "        if p:\n",
    "            print('Number of rounds: ' + str(n))\n",
    "            print('Miscommunication chance: ' + str(m) + '%')\n",
    "            print(self.player0.name + ' score: ' + str(player0Score))\n",
    "            print(self.player1.name + ' score: ' + str(player1Score))\n",
    "        return [player0Score,player1Score]"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {
    "collapsed": false
   },
   "source": [
    "Nice player always cooperates."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 2,
   "metadata": {
    "collapsed": false
   },
   "outputs": [
   ],
   "source": [
    "class nicePlayer:\n",
    "    def __init__(self):\n",
    "        self.name = 'Nice Player'\n",
    "    def decision(self,n,decisions):\n",
    "        return 1"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {
    "collapsed": false
   },
   "source": [
    "Mean player always cheats."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 3,
   "metadata": {
    "collapsed": false
   },
   "outputs": [
   ],
   "source": [
    "class meanPlayer:\n",
    "    def __init__(self):\n",
    "        self.name = 'Mean Player'\n",
    "    def decision(self,n,decisions):\n",
    "        return 0"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {
    "collapsed": false
   },
   "source": [
    "Grudge player starts by cooperating, but if you cheat, it will cheat back forever."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 4,
   "metadata": {
    "collapsed": false
   },
   "outputs": [
   ],
   "source": [
    "class grudgePlayer:\n",
    "    def __init__(self):\n",
    "        self.name = 'Grudge Player'\n",
    "    def decision(self,n,decisions):\n",
    "        if 0 in [b for a,b in decisions]:\n",
    "            return 0\n",
    "        else:\n",
    "            return 1"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {
    "collapsed": false
   },
   "source": [
    "Copycat player starts by cooperating and then repeats whatever the other player did last."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 5,
   "metadata": {
    "collapsed": false
   },
   "outputs": [
   ],
   "source": [
    "class copycatPlayer:\n",
    "    def __init__(self):\n",
    "        self.name = 'Copycat Player'\n",
    "    def decision(self,n,decisions):\n",
    "        if len(decisions):\n",
    "            return decisions[-1][int(not n)]\n",
    "        else:\n",
    "            return 1"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {
    "collapsed": false
   },
   "source": [
    "Forgiving copycat player starts by cooperating and then cheats back only if cheated twice in a row."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 6,
   "metadata": {
    "collapsed": false
   },
   "outputs": [
   ],
   "source": [
    "class forgivingCopycatPlayer:\n",
    "    def __init__(self):\n",
    "        self.name = 'Forgiving Copycat Player'\n",
    "    def decision(self,n,decisions):\n",
    "        if len(decisions)>1 and (decisions[-1][int(not n)] == 0 and decisions[-2][int(not n)] == 0):\n",
    "            return 0\n",
    "        else:\n",
    "            return 1"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {
    "collapsed": false
   },
   "source": [
    "Simple copycat player starts by cooperating. If you cheat, it does the opposite of what it did last. If you cooperate, it repeats whatever it did last."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 7,
   "metadata": {
    "collapsed": false
   },
   "outputs": [
   ],
   "source": [
    "class simpleCopycatPlayer:\n",
    "    def __init__(self):\n",
    "        self.name = 'Simple Copycat Player'\n",
    "    def decision(self,n,decisions):\n",
    "        if len(decisions):\n",
    "            if decisions[-1][int(not n)]:\n",
    "                return decisions[-1][n]\n",
    "            else:\n",
    "                return int(not decisions[-1][n])\n",
    "        else:\n",
    "            return 1"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 8,
   "metadata": {
    "collapsed": false
   },
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "Number of rounds: 10\n",
      "Miscommunication chance: 5%\n",
      "Copycat Player score: 21\n",
      "Forgiving Copycat Player score: 17\n"
     ]
    },
    {
     "data": {
      "text/plain": [
       "[21, 17]"
      ]
     },
     "execution_count": 8,
     "metadata": {
     },
     "output_type": "execute_result"
    }
   ],
   "source": [
    "game(copycatPlayer(),forgivingCopycatPlayer()).play(m=5)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 15,
   "metadata": {
    "collapsed": false
   },
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "['Mean Player', 465]\n",
      "['Mean Player', 465]\n",
      "['Mean Player', 465]\n",
      "['Mean Player', 465]\n",
      "['Mean Player', 465]\n",
      "['Copycat Player', 375]\n",
      "['Copycat Player', 375]\n",
      "['Copycat Player', 375]\n",
      "['Copycat Player', 375]\n",
      "['Copycat Player', 375]\n",
      "['Nice Player', 330]\n",
      "['Nice Player', 330]\n",
      "['Nice Player', 330]\n",
      "['Nice Player', 330]\n",
      "['Nice Player', 330]\n",
      "['Nice Player', 330]\n",
      "['Nice Player', 330]\n",
      "['Nice Player', 330]\n",
      "['Nice Player', 330]\n",
      "['Nice Player', 330]\n",
      "['Nice Player', 330]\n",
      "['Nice Player', 330]\n",
      "['Nice Player', 330]\n",
      "['Nice Player', 330]\n",
      "['Nice Player', 330]\n"
     ]
    }
   ],
   "source": [
    "players = [copycatPlayer(),copycatPlayer(),copycatPlayer(),copycatPlayer(),copycatPlayer(),\n",
    "           meanPlayer(),meanPlayer(),meanPlayer(),meanPlayer(),meanPlayer(),\n",
    "           nicePlayer(),nicePlayer(),nicePlayer(),nicePlayer(),nicePlayer(),\n",
    "           nicePlayer(),nicePlayer(),nicePlayer(),nicePlayer(),nicePlayer(),\n",
    "           nicePlayer(),nicePlayer(),nicePlayer(),nicePlayer(),nicePlayer(),]\n",
    "N = len(players)\n",
    "scores = [0 for i in range(N)]\n",
    "M = 0\n",
    "\n",
    "for i in range(N-1):\n",
    "    for j in range(i+1,N):\n",
    "        a,b = game(players[i],players[j]).play(m=M,p=0)\n",
    "        scores[i] += a\n",
    "        scores[j] += b\n",
    "for output in [[name.name,score] for name,score in sorted(zip(players,scores),key=lambda x: x[1])[::-1]]:\n",
    "    print(output)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 0,
   "metadata": {
    "collapsed": false
   },
   "outputs": [
   ],
   "source": [
   ]
  }
 ],
 "metadata": {
  "kernelspec": {
   "display_name": "Python 2 (Ubuntu Linux)",
   "language": "python",
   "name": "python2-ubuntu"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 2
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython2",
   "version": "2.7.15rc1"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 0
}